/* -*- Mode: C++ -*- */

/* COPYING
 * Liquid Rendering Library
 * (c) 2009 by Henrik Theiling
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * SYNOPSIS
 * Automatical joining based on existing glyph variants.  This does not
 * use arabicshaping.txt, but checks for join-bits in the font.  Otherwise,
 * it's the Unicode algorithm.
 *
 * ZWJ and ZWNJ are handled in liquid.c already.  We will skip any glyphs
 * that have LIQ_FLAG_OVERRIDE_LEFT/RIGHT set.  Ligatures should be
 * inserted before this algorithm is run, because ligatures may have
 * joining behaviour (e.g. Arabic lam-alef).
 *
 * NOTE: The bidi algorithm deletes BN characters, which is bad for this
 *       algorithm.  That's why it sets the bit 'LIQ_FLAG_TRANSPARENT'
 *       before starting to delete characters.
 */

#include "internal.h"

static void set_left_right (
    LiqCharInfo     **left,
    LiqGlyphFlagSet  *left_flags,
    LiqFont const    *font,
    LiqCharInfo      *right)
{
    if (right->important && right->opaque) {
        LiqGlyphFlagSet right_flags= 0;
        LiqGlyph        glyph= font->first_glyph (right->code);
        if (glyph) {
            /* gather flags: */
            for (; glyph; glyph= font->next_glyph(glyph))
                right_flags|= font->glyph_flags (glyph);

            /* check whether joining is possible: */
            if ((*left_flags & LIQ_GLYPH_JOIN_RIGHT) && (right_flags & LIQ_GLYPH_JOIN_LEFT)) {
                /* unless overridden, set left joining behaviour: */
                unless ((*left)->joining & LIQ_JOIN_MASK_RIGHT)
                    (*left)->joining|= LIQ_JOIN_RIGHT;

                /* unless overridden, set right joining behaviour: */
                unless (right->joining & LIQ_JOIN_MASK_LEFT)
                    right->joining|= LIQ_JOIN_LEFT;
            }
        }

        *left_flags= right_flags;
        *left=       right;
    }
}

void liq_run_auto_join (
    LiqRender *self,
    LiqParagraphInfo *para,
    int run_level,
    LiqBidiType sor,
    LiqBidiType eor,
    LiqCharInfo *first,
    LiqCharInfo *last)
{
    LiqFont const *const font= self->m_font;
    LiqCharInfo    *prev;
    LiqCharInfo    *cur;
    LiqGlyphFlagSet prev_flags;

    (void)para;
    (void)sor;
    (void)eor;

    /* This algorithm runs when the sequence is still in logical order,
     * but the glyph composition is in visual order.  So we swap the
     * processing direction depending on run_level. */
    prev= NULL;
    prev_flags= 0;
    if (is_odd (run_level)) /* right-to-left */
        for (cur= last; cur >= first; --cur)
            set_left_right (&prev, &prev_flags, font, cur);
    else                    /* left-to-right */
        for (cur= first; cur <= last; ++cur)
            set_left_right (&prev, &prev_flags, font, cur);
}

