/* -*- Mode: C++ -*- */

/* COPYING
 * Liquid Rendering Library
 * (c) 2009 by Henrik Theiling
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * SYNOPSIS
 * Bidi algorithm, according to Unicode 5.1.0, Technical Report 9.
 *
 * Changes/Comments:
 *    - X8: embedding of paragraph separators:
 *      This is performed implicitly (in liq_split_paragraphs) by
 *      excluding the paragraph separators from the bidi algorithm
 *      altogether, i.e., they are not part of a paragraph, but
 *      they are inbetween.
 *
 *    - L3: revering NSMs after logical -> graphical reordering
 *      This is not done. We want stray NSMs on the left in R2L mode.
 *
 *    - L4: mirror glyph selection
 *      This is done earlier, namely before all Lx rules, because
 *      we need to know the exact glyph before automatic line
 *      breaking in order to do kerning correctly and thus know the
 *      width in pixels of each glyph exactly.
 */

#include "internal.h"

#ifdef LIQ_DEBUG

static char const *bidi2str (LiqBidiType b)
{
    switch (b) {
    case LIQ_BIDI_NUMBER:
    case LIQ_BIDI_STRONG_L:
    case LIQ_BIDI_STRONG_R:
    case LIQ_BIDI_STRONG_LR:
    case LIQ_BIDI_SEP_TERM:
                       return "ERR";
    case LIQ_BIDI_L:   return "L";
    case LIQ_BIDI_R:   return "R";
    case LIQ_BIDI_AL:  return "AL";
    case LIQ_BIDI_LRE: return "LRE";
    case LIQ_BIDI_LRO: return "LRO";
    case LIQ_BIDI_RLE: return "RLE";
    case LIQ_BIDI_RLO: return "RLO";
    case LIQ_BIDI_EN:  return "EN";
    case LIQ_BIDI_AN:  return "AN";
    case LIQ_BIDI_ES:  return "ES";
    case LIQ_BIDI_ET:  return "ET";
    case LIQ_BIDI_CS:  return "CS";
    case LIQ_BIDI_PDF: return "PDF";
    case LIQ_BIDI_BN:  return "BN";
    case LIQ_BIDI_NSM: return "NSM";
    case LIQ_BIDI_ON:  return "ON";
    case LIQ_BIDI_B:   return "B";
    case LIQ_BIDI_S:   return "S";
    case LIQ_BIDI_WS:  return "WS";
    }
    return "ERR";
}

static LIQ_UNUSED
void show_char_info(
    char const *step,
    int level, LiqBidiType sor, LiqBidiType eor,
    LiqCharInfo const *first, LiqCharInfo const *last)
{
    LiqCharInfo const *cur;
    fprintf (stderr, "DEBUG: %s: %2d (%-3s)", step, level, bidi2str(sor));
    for (cur=first; cur <= last; ++cur)
        fprintf (stderr, " %d%-3s", cur->level, bidi2str(cur->bidi));
    fprintf (stderr, " (%-3s)\n", bidi2str(eor));
}

#else
#define show_char_info(A,B,C,D,E,F)
#endif

/* ******************************************************************************** */
/* BIDI category without NSM: generated list for bsearch(): */

typedef struct _BidiRange {
    LIQ_UBITS32 start : 32; /* keep this word aligned to make access faster */
    LIQ_UBITS32 end   : 24; /* minimum 21, because we don't know signedness */
    LIQ_UBITS8  bidi  : 8;
} BidiRange;

static BidiRange const bidi_range[]= {
#include "autogen/bidi2.inc"
};

static size_t const bidi_range_cnt= sizeof(bidi_range) / sizeof(bidi_range[0]);

static int bidi_cmp (void const *a, void const *b)
{
    LiqChar32 key=             *(LiqChar32 const *)a;
    BidiRange const *arr_elem=  (BidiRange const *)b;
    return key < (LiqChar32)arr_elem->start ?
               -1
           : key > (LiqChar32)arr_elem->end ?
               +1
           :   0;
}

static LiqBidiType liq_get_bidi_type (LiqChar32 c)
{
    BidiRange *range=
        (BidiRange*)bsearch (
            &c,
            bidi_range,
            bidi_range_cnt,
            sizeof(bidi_range[0]),
            bidi_cmp);
    if (range)
        return range->bidi;
    return LIQ_BIDI_L;
}

/* ******************************************************************************** */
/* Mirroring: mkhashtable generated table */

typedef struct _BidiMirror {
    LiqUInt32 orig;
    LiqChar32 mirror; /* if 0, no mirror codepoint available */
} BidiMirror;
typedef BidiMirror liq_mirror_map_t;

#define LIQ_MIRROR_MAP_KEY_MATCH(p,c) ((p)->orig == (c))

#include "autogen/mirror.inc"

/* ********************************************************************** */
/* Exception list for transparency (we use NSM and BN to preselect) */

typedef struct _BidiTrans {
    LIQ_UBITS32 code        : 21;
    LIQ_UBITS8  transparent : 1;
} BidiTrans;
typedef BidiTrans liq_trans_map_t;

#define LIQ_TRANS_MAP_KEY_MATCH(p,c) ((p)->code == (c))

#include "autogen/trans.inc"

static LiqBool liq_char_is_transparent (LiqCharInfo *ci)
{
    BidiTrans const *m= liq_trans_lookup(ci->code);
    if (m)
        return m->transparent; /* exception from our simple rule */
    return ci->bidi == LIQ_BIDI_NSM ||
           ci->bidi == LIQ_BIDI_BN;
}

/* ********************************************************************** */
/* In this paragraph, the Bidirectional (Bidi) Algorithm is implemented.
 * The rule tags in the code refer to Unicode Tr9. */

typedef struct _XStatus {
    int         level;
    LiqBidiType override;
} XStatus;

typedef struct _XStack {
    XStatus mem[64];
    XStatus *ptr;
} XStack;

static int find_para_level(LiqCharInfo *first, LiqCharInfo *last)
{
    LiqCharInfo *cur;
    for (cur= first; cur <= last; ++cur)
        if (cur->bidi & LIQ_BIDI_STRONG_LR)                                          /* P2 */
            return cur->bidi & LIQ_BIDI_STRONG_R ? 1 : 0;                            /* P3 */
    return 0;                                                                        /* P3 */
}

static LiqBool push_embed_level (XStack *stack, int new_level)
{
    unless (new_level <= 61)
        return LIQ_FALSE;

    stack->ptr[1]= stack->ptr[0];
    stack->ptr++;
    stack->ptr->level= new_level;
    return LIQ_TRUE;
}

static LiqCharInfo *liq_bidi_resolve_explicit (
    int para_level, LiqCharInfo *first, LiqCharInfo *last)
{
    XStack xstack;
    int stack_overflow= 0;
    LiqCharInfo *src= first;
    LiqCharInfo *dst= first; /* X9: we copy all but special while running the algorithm */

    /* Remaining: L R AL LRE LRO RLE RLO | EN AN ES ET CS PDF BN NSM | B S WS ON */

    /* 3.3.2 */
    /* X1 */
    xstack.ptr= xstack.mem;
    xstack.ptr->level= para_level;
    xstack.ptr->override= LIQ_BIDI_NEUTRAL;
    for (; src <= last; ++src) {
        switch (src->bidi) {
        case LIQ_BIDI_RLE:                                                           /* X2 */
            unless (push_embed_level (&xstack, (xstack.ptr->level + 1) | 1))
                ++stack_overflow;
            break;

        case LIQ_BIDI_LRE:                                                           /* X3 */
            unless (push_embed_level (&xstack, (xstack.ptr->level | 1) + 1))
                ++stack_overflow;
            break;

        case LIQ_BIDI_RLO:                                                           /* X4 */
            if (push_embed_level (&xstack, (xstack.ptr->level + 1) | 1))
                xstack.ptr->override= LIQ_BIDI_R;
            else
                ++stack_overflow;
            break;

        case LIQ_BIDI_LRO:                                                           /* X5 */
            if (push_embed_level (&xstack, (xstack.ptr->level | 1) + 1))
                xstack.ptr->override= LIQ_BIDI_L;
            else
                ++stack_overflow;
            break;

        default:                                                                     /* X6 */
            src->level= xstack.ptr->level;
            if (xstack.ptr->override != LIQ_BIDI_NEUTRAL)
                src->bidi= xstack.ptr->override;
            *dst++= *src;                                        /* X9: we copy on the fly */
            break;

        case LIQ_BIDI_PDF:                                                           /* X7 */
            if (stack_overflow)
                --stack_overflow;
            else
            if (xstack.ptr > xstack.mem)
                xstack.ptr--;
            break;

        case LIQ_BIDI_BN:                                                            /* X9 */
            break;
        }
    }

    /* Remaining: L R AL | EN AN ES ET CS NSM | B S WS ON */

    /* src is used to overwrite all charinfos from dst..last */
    for (src= dst; src <= last; ++src)                  /* we move all X9 codes to the end */
        *src= liq_char_info_0;

    return dst-1;
}

static void liq_bidi_resolve_weak (
    LiqBidiType sor, LiqCharInfo *first, LiqCharInfo *last)
{
    LiqBidiType prev_bidi=   sor;
    LiqBidiType prev_strong= sor;
    LiqCharInfo *cur;

    /* Remaining: L R AL | EN AN ES ET CS NSM | B S WS ON */

    /* W1, W2, W3 */
    for (cur= first; cur <= last; ++cur) {
        switch (cur->bidi) {
        default: break;

        case LIQ_BIDI_NSM:                                                           /* W1 */
            cur->bidi= prev_bidi;
            break;

        case LIQ_BIDI_L:
        case LIQ_BIDI_R:
            prev_strong= cur->bidi;                                                  /* W2 def */
            break;

        case LIQ_BIDI_AL:
            prev_strong= cur->bidi;                                                  /* W2 def */
            cur->bidi= LIQ_BIDI_R;                                                   /* W3 */
                /* remember AL as AL, but change it to R on the fly */
            break;

        case LIQ_BIDI_EN:                                                            /* W2 use */
            if (prev_strong == LIQ_BIDI_AL)
                cur->bidi= LIQ_BIDI_AN;
            break;
        }

        prev_bidi= cur->bidi;
    }
    /* Remaining: L R | EN AN ES ET CS | B S WS ON */

    /* We split here and restart the loop, because W2 influences W4, e.g.:
     * AL EN CS EN --W2-> AL AN CS AN --W4-> AL AN AN AN
     *
     * We could implement this in one go, but it's easier and clearer
     * to use two loops.
     */

    /* W4, W5 */
    prev_bidi= sor;
    for (cur= first; cur <= last; ++cur) {
        switch (cur->bidi) {
        default: break;

        case LIQ_BIDI_ES:                                                            /* W4.a */
            if (cur < last && prev_bidi == LIQ_BIDI_EN && cur[1].bidi == LIQ_BIDI_EN)
                cur->bidi= LIQ_BIDI_EN;
            break;

        case LIQ_BIDI_CS:                                                            /* W4.b */
            if (cur < last) {
                if ((prev_bidi & LIQ_BIDI_NUMBER) && prev_bidi == cur[1].bidi)
                    cur->bidi= prev_bidi;
            }
            break;

        case LIQ_BIDI_ET:                                                            /* W5 <- */
            if (prev_bidi == LIQ_BIDI_EN)
                cur->bidi= LIQ_BIDI_EN; /* this handles sequences, to */
            break;

        case LIQ_BIDI_EN: {                                                          /* W5 -> */
                LiqCharInfo *cur2;
                for (cur2= cur-1; cur2 >= first && cur2->bidi == LIQ_BIDI_ET; --cur2)
                    cur2->bidi= LIQ_BIDI_EN;
            }
            break;
        }

        prev_bidi= cur->bidi; /* remember EN as EN, not as L */
    }

    /* W6, W7: separate because W5 search both directions, which is only complete now. */
    prev_strong= sor;
    for (cur= first; cur <= last; ++cur) {
        if (cur->bidi & LIQ_BIDI_SEP_TERM)                                      /* W6 */
            cur->bidi= LIQ_BIDI_ON;
        else
        if (cur->bidi == LIQ_BIDI_EN && prev_strong == LIQ_BIDI_L)              /* W7 use */
            cur->bidi= LIQ_BIDI_L;
        else
        if (cur->bidi & LIQ_BIDI_STRONG)                                        /* W7 def */
            prev_strong= cur->bidi;
    }

    /* Remaining: L R | EN AN | B S WS ON */
}

static void liq_bidi_resolve_neutral (
    LiqBidiType embed_type, LiqBidiType sor, LiqBidiType eor, LiqCharInfo *first, LiqCharInfo *last)
{
    LiqBidiType prev_non_neutral;
    LiqCharInfo *cur;

    /* Remaining: L R | EN AN | B S WS ON */

    /* backwards to mark L/R back context: */
    prev_non_neutral= eor;
    for (cur= last; cur >= first; --cur) {
        if (prev_non_neutral == LIQ_BIDI_L) {
            assert (!cur->mirror);
            cur->mirror= LIQ_TRUE;   /* all others (R, EN, AN) count as R */
        }

        if (cur->bidi & (LIQ_BIDI_STRONG | LIQ_BIDI_WEAK))
            prev_non_neutral= cur->bidi; /* L, R, EN, AN */
    }

    /* forwards to check whether context is the same */
    prev_non_neutral= sor;
    for (cur= first; cur <= last; ++cur) {
        if (cur->bidi & (LIQ_BIDI_STRONG | LIQ_BIDI_WEAK))
            prev_non_neutral= cur->bidi;
        else {
            if (prev_non_neutral == LIQ_BIDI_L && cur->mirror)                         /* N1 */
                cur->bidi= LIQ_BIDI_L;
            else
            if (prev_non_neutral != LIQ_BIDI_L && !cur->mirror)
                cur->bidi= LIQ_BIDI_R;
            else
                cur->bidi= embed_type;                                                 /* N2 */
        }
    }

    /* remove the MIRROR flags again */
    for (cur= last; cur >= first; --cur)
        cur->mirror= LIQ_FALSE;

    /* Remaining: L R | EN AN */
}

static void liq_bidi_resolve_implicit (LiqCharInfo *first, LiqCharInfo *last)
{
    /* Remaining: L R | EN AN */

    /* I1, I2 */
    LiqCharInfo *cur;
    for (cur= first; cur <= last; ++cur) {
        switch (cur->bidi) {
        case LIQ_BIDI_L:
            cur->level= (cur->level + 1) & ~1; /* equal or next even */
            break;

        case LIQ_BIDI_R:
            cur->level= cur->level | 1;        /* equal or next odd */
            break;

        default: /* LIQ_BIDI_EN, LIQ_BIDI_AN */
            cur->level= (cur->level | 1) + 1;  /* strictly greater, next odd */
            break;
        }
    }
}

static void liq_bidi_resolve_rest (
    LiqRender    *self,
    LiqParagraphInfo *para,
    int run_level, LiqBidiType sor, LiqBidiType eor, LiqCharInfo *first, LiqCharInfo *last)
{
    (void)self;
    (void)para;

    liq_bidi_resolve_weak (sor, first, last);
    liq_bidi_resolve_neutral (level2bidi(run_level), sor, eor, first, last);
    liq_bidi_resolve_implicit (first, last);
}

static void liq_bidi_reverse_contiguous (int min_level, LiqCharInfo *first, LiqCharInfo *last)
{
    LiqCharInfo *cur;
    LiqCharInfo *begin= NULL;

    /* invoke liq_bidi_reverse for contiguous regions of level >= min_level */
    for (cur=first; cur <= last; ++cur) {
        if (cur->bidi != LIQ_BIDI_BN &&
            cur->level < min_level) /* end of region */
        {
            if (begin) {
                liq_seq_reverse (begin, cur-1);
                begin= NULL;
            }
        }
        else
        if (!begin)
            begin= cur; /* beginning of region */
    }
    if (begin)
        liq_seq_reverse (begin, last);
}

#if 0
/* L3: Reverse NSMs: Not done, in Hebrew, Arabic, etc., stray NSMs will be the left. */
static void liq_bidi_reverse_nsm(LiqCharInfo *first, LiqCharInfo *last)
{
    LiqCharInfo *first_nsm;
    LiqCharInfo *cur;

    assert (first <= last);

    first_nsm= last; /* start with an empty range */
    for (cur=first; cur <= last; ++cur) {
        assert (cur->orig_bidi != LIQ_BIDI_BN); /* must have been removed */
        if (cur->orig_bidi == LIQ_BIDI_NSM && is_odd(cur->level)) {
            unless (first_nsm)
                first_nsm= cur;
        }
        else
        if (first_nsm < cur) {
            liq_seq_reverse (first_nsm, cur);
            first_nsm= last; /* empty range again */
        }
    }
}
#endif

static void liq_bidi_line (LiqLineInfo *line)
{
    int i;
    LiqParagraphInfo *const para= line->paragraph;
    LiqCharInfo *const first= line->first;
    LiqCharInfo *const last=  line->last;
    int max_level=      0;
    int min_odd_level= 99; /* this is larger than anything that may occur */
    LiqBool eos= LIQ_TRUE;                                                       /* L1.4 set */
    LiqCharInfo  *cur;

    assert (first <= last);

    for (cur=last; cur >= first; --cur) {
        if (cur->bidi == LIQ_BIDI_BN)
            continue;

        /* L1 */
        if (cur->orig_bidi & LIQ_BIDI_B_OR_S) {                                /* L1.1, L1.2 */
            eos= LIQ_TRUE;                                                       /* L1.3 set */
            cur->level= para->level;
        }
        else
        if (eos && (cur->orig_bidi == LIQ_BIDI_WS))                        /* L1.3, L1.4 use */
            cur->level= para->level;
        else
            eos= LIQ_FALSE;                                                      /* L1.3 set */

        /* L2 set */
        if (cur->level > max_level)
            max_level= cur->level;
        if (is_odd(cur->level) && cur->level < min_odd_level)
            min_odd_level= cur->level;
    }

    /* L2 use */
    for (i=max_level; i >= min_odd_level; i--)
        liq_bidi_reverse_contiguous (i, first, last);

    /* L3: Reverse NSMs.
     *     - after swapping, in all odd areas, reverse base+NSMs:
     *       [NSM .. NSM BASE] -> [BASE NSM ... NSM]
     *       because these sequences where swapped by L2.
     *
     * Not done, see comments above and below about L3.
     */
    /* liq_bidi_reverse_nsm(first, last); */
}

static void liq_bidi_mirror_codepoints (LiqRender *self, LiqCharInfo *first, LiqCharInfo *last)
{
    LiqCharInfo *cur;
    LiqFont const *const font= self->m_font;

    /* find mirrored codepoints or mark as needing manual mirrroring: */
    for (cur= first; cur <= last; ++cur) {
        if (is_odd(cur->level)) {
            BidiMirror const *m= liq_mirror_lookup (cur->code);
            assert (!cur->is_glyph);
            if (m) { /* needs mirroring? */
                if (m->mirror && font->first_glyph (m->mirror))
                    cur->code= m->mirror;              /* change code to mirrored codepoint */
                else
                    cur->mirror= LIQ_TRUE;             /* mirror manually */
            }
        }
    }
}

static void liq_bidi_swap_explicit_joiners (LiqCharInfo *first, LiqCharInfo *last)
{
    LiqCharInfo *cur;
    for(cur= first; cur <= last; ++cur)
        if (is_odd(cur->level))
            cur->joining= liq_joining_swap_left_right(cur->joining);
}

static void liq_bidi_paragraph (LiqRender *self, LiqParagraphInfo *para)
{
    assert (para->first <= para->last);

    /* P2,P3 */
    if (para->level < 0)
        para->level= find_para_level (para->first, para->last);

    /* X */
    para->last= liq_bidi_resolve_explicit (para->level, para->first, para->last);
    unless (para->first <= para->last) /* X9 might shrink paragraph to 0 */
        return;

    /* X10 */
    liq_foreach_run (self, para, &liq_bidi_resolve_rest);

    /* switch order of ZWJ and ZWNJ settings for right-to-left runs. */
    liq_bidi_swap_explicit_joiners (para->first, para->last);

    /* set joining (before ligature selection) */
    liq_foreach_run (self, para, &liq_run_auto_join);

    /* code point selection (ligatures) */
    liq_para_select_ligatures (self, para);

    /* L4: mirroring */
    liq_bidi_mirror_codepoints (self, para->first, para->last);

    /* glyph selection */
    liq_para_select_glyphs (self, para);

    /* FIXME: We can insert automatic line breaking here.  See Tr9 3.4.
     *        We move L4 and do it early now, before glyph selection,
     *        so that we can compute the exact widths of all characters,
     *        kerning inclusive.
     */
}


static void liq_bidi_remember_join_state (LiqCharInfo *first, LiqCharInfo *last)
{
    LiqCharInfo *cur;
    LiqCharInfo *prev;
    LiqJoining   left_join= LIQ_NO_JOIN;

    /* We need to swap LEFT and RIGHT afterwards for right-to-left runs,
     * because the joiners are processed in logical order, while they
     * affect the glyphs in visual order.
     *
     * Also note how this handles sequences of ZWJ and ZWNJ: ZWNJ will
     * cause the NO_JOIN bits to be set across ZWJers, while the ZWJ
     * will not work across ZWNJ.  This is to handle sequences like
     * ZWJ ZWNJ ZWJ, which should cause both glyphs to have both flags
     * set (JOIN and NO_JOIN), indicating joining without ligation.
     *
     * As a side effect, X ZWJ ZWNJ Y will leave X with JOIN|NO_JOIN,
     * and y with NO_JOIN.  That's ok, because in this situation, no
     * joining and no ligation will be done, X will select joined glyph,
     * Y will select non-joined glyph.  Fine.
     */
    prev= NULL;
    for (cur= first; cur <= last; ++cur) {
        switch (cur->code) {
        case UNICODE_ZWJ:
            left_join|= LIQ_JOIN_LEFT;                          /* keep preceding ZWNJ setting */
            if (prev) {
                prev->joining|= LIQ_JOIN_RIGHT;
                                  /* do not reset 'prev' to allow succeeding ZWNJ to change it */
            }
            break;

        case UNICODE_ZWNJ:
            left_join= LIQ_NO_JOIN_LEFT;        /* prevent preceding ZWJ to change right glyph */
            if (prev) {
                prev->joining|= LIQ_NO_JOIN_RIGHT;
                prev= NULL;                     /* prevent succeeding ZWJ to change left glyph */
            }
            break;

        default:
            if (cur->important && cur->opaque) {
                cur->joining|= left_join;
                left_join= LIQ_NO_JOIN;
                prev= cur;
                break;
            }
        }
    }
}

static void liq_bidi_set_bidi_types (LiqCharInfo *first, LiqCharInfo *last)
{
    LiqCharInfo *cur;
    for (cur=first; cur <= last; ++cur) {
        if (cur->bidi == LIQ_BIDI_ON) /* yet unset? */
            cur->bidi= liq_get_bidi_type(cur->code);
        cur->orig_bidi= (cur->bidi <= LIQ_ORIG_BIDI_MAX) ? cur->bidi : LIQ_BIDI_ON;

        cur->opaque= !liq_char_is_transparent (cur);
    }
}

void liq_render_bidi (LiqRender *self)
{
    int i;
    assert (self && self->m_info && "expected liq_render_set_text() first");
    assert (self && self->m_font && "expected liq_render_set_font() first");

    if (self->m_bidi) /* already done. */
        return;

    /* get bidi types */
    liq_bidi_set_bidi_types (self->m_info, self->m_info + self->m_info_cnt - 1);

#if 0
    show_char_info ("A", self->m_para_level, LIQ_BIDI_ON, LIQ_BIDI_ON,
        self->m_info, self->m_info + self->m_info_cnt - 1);
#endif

    /* Preparation: remember states from ZWJ and ZWNJ */
    liq_bidi_remember_join_state (self->m_info, self->m_info + self->m_info_cnt - 1);

    /* P1, we let the separators untouched: they are neither in the previous, nor
     *     in the next paragraph, but in between.  This is because otherwise we'd
     *     have to check the 'code' of a char to determine whether its a paragraph
     *     separator in rule X8.
     *     Since they are not included in the algorithm, there is no
     *     implementation for rule X8 here.
     */
    for (i=0; i < self->m_paragraph_cnt; i++)
        liq_bidi_paragraph (self, self->m_paragraph + i);

    /* L1,L2
     *
     * L3: Not done, we'll render stray NSMs on the left side in Arabic mode.
     *     This seems more logical, because it's consistently in reading
     *     order this way.
     *
     * L4: Done earlier in bidi_paragraph, because mirroring is vital for
     *     glyph selection and kerning.
     */
    liq_split_lines (self);
    for (i=0; i < self->m_line_cnt; i++)
        liq_bidi_line (self->m_line + i);

    self->m_bidi= LIQ_TRUE;

#if 0
    show_char_info ("Z", self->m_para_level, LIQ_BIDI_ON, LIQ_BIDI_ON,
        self->m_info, self->m_info + self->m_info_cnt - 1);
#endif
}

