/* -*- Mode: C -*- */

/* COPYING
 * Liquid Rendering Library
 * (c) 2009 by Henrik Theiling
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef LIQUID_FONT_H_239843517923124
#define LIQUID_FONT_H_239843517923124

#include <liquid/render.h>

#ifdef __cplusplus
extern "C" {
#endif

/* FONT API
 **********
 * This describes version 1 of the font API.
 *
 * You only need this if you want to provide your own fonts.
 *
 * The font interface was implemented in an abstract way so that the exact storage of pixel
 * data and additional information in the font is not specified nor enforced nor exposed
 * in any way.  For a font, you only need the LiqFont structure and can then set up
 * its 'member' functions as follows:
 *
 *    first_glyph     - maps a code point to its first glyph.  Returns 0 if none was found.
 *                      This maps to a range, because there might be glyph variants.  To iterate
 *                      the glyphs, there is next_glyph().
 *                      This library chooses one of the glyphs by several criteria.
 *                      The default glyph is the first glyph.
 *                      Note: it is ok to access the glyph behind the last one ever returned
 *                      here, i.e., there MUST be a dummy, all-zero sentinel glyph at the
 *                      very end of the glyph list.
 *
 *                      Note than in monospaced mode, a glyph may still be wider than a
 *                      LiqFont::width.  It will then be displayed on several multiples
 *                      of LiqFont::width.  This may happen for extrawide glyphs, e.g.
 *                      ideographic characters may need two positions while Latin
 *                      characters only occupy one.  Font designers should remember this
 *                      when centering wider glyphs for monospace display.
 *
 *    next_glyph      - returns the next variant glyph.  Returns 0 if there is none.
 *                      If a glyph is explicitly requested by the user by a Unicode
 *                      variant selector, that glyph is used by using its position
 *                      as its variation number: the glyphs are implicitly enumerated
 *                      started with variant 0.
 *                      NOTE: you cannot simply increment the glyph ID by one to get
 *                      the next glyph because a font may use more than one (internal)
 *                      glyph ID to store larger bitmaps.
 *
 *    glyph_x1        - maps a glyph id to the x coordinate of first pixel column.
 *    glyph_x2        - maps a glyph id to the x coordinate of last pixel column.
 *                      For mono-spaced mode, these coordinates are not used.  Instead,
 *                      0..LiqFont::bitwidth-1 is used.
 *                      x-coordinates are left-to-right (left-most bit at index x1 (or 0 for mono))
 *
 *    glyph_data      - maps a glyph id and the column number to the pixel data.
 *                      Column bits are top-to-bottom (top-most bit = 2^0 = 1).
 *
 *    glyph_attr      - maps a glyph id to its glyph attribute structure, needed for
 *                      automatic glyph selection.
 *
 *    glyph_space_col - These pixels are used to insert spacing pixel columns.  May be
 *                      different from 0 for cursive scripts where a linked line is
 *                      needed (-> U+0640 tatweel).  For a pair of two glyphs L and R
 *                      (L=left, R=right, in graphical order), the space_col of L will
 *                      be used to fill the space between L and R.  If != 0, it will
 *                      be drawn until a pixel is hit on each like of the space.  The
 *                      space_col will not be drawn into R if all pixel columns
 *                      intersected with space_col are empty.  There is no such
 *                      constraint for L (because for L, you can defined space_col
 *                      explicitly, while for R, L's space_col is used).
 *
 *    glyph_space_x   - Amount of spacing pixel columns on the visual right of the glyph
 *                      except at ends of lines.  Typically 0 for block graphics,
 *                      box drawing, etc., and for some types of space.
 *
 *    glyph_flags     - see LiqGlyphFlags
 *
 *    Glyphs are selected by the library from first to last glyph.  The first best
 *    match will be used.
 *
 * Ligature table:
 *    compose         - Maps two adjacent code points (in visual order) to a ligature.
 *                      Returns 0 if none is found (the Unicode NUL character is not a
 *                      ligature).  The additional boolean is LIQ_TRUE if there was a
 *                      ZWJ between the two original code points, while LIQ_FALSE signifies
 *                      that there was none.  This can be used to distinguish optional
 *                      ligatures from mandatory ones (e.g. optional: Hebrew LAMED+ALEPH,
 *                      mandatory: Arabic LAM+ALIPH).  With a ZWJ, the optional ligatures
 *                      should be returned, while without, they should not be returned.
 *
 *                      Ligatures are composed recursively, one-by-one from left-to-right
 *                      in visual order.  NSMs are *NOT* skipped, because if a combination
 *                      of base + combining is unavailable, this library has no way to
 *                      emulate diacritic placement -- stray NSMs are displayed linearly.
 *                      So the output would probably be worse if base1 + dia1 + base1 is
 *                      transformed to ligature + dia1 when dia1 remains stray.
 *                      For this to be nice, the font must support base1+dia1 as well
 *                      as ligature+dia1, and if both are present, nice display is still
 *                      possible.
 *
 *                      If a codepoint returned here cannot be found in the font, it
 *                      will not be used, so you can insert 'dummy' values if your
 *                      font can compose a+b+c, but not a+b.  For ab, just use a
 *                      dummy codepoint and compose ab+c to abc.  If there is no c,
 *                      a and b will still remain separate.
 *
 *                      base + NSM*, as specified as canonical equivalence in Unicode,
 *                      will automatically be composed if present in the font.  Only
 *                      compatibility compositions need to be managed by the font,
 *                      because we do cannot easily know whether they are optional
 *                      or mandatory.  Of course, you may have more ligatures
 *                      than present in Unicode presentation forms.
 *
 *                      Example of invocation order:
 *                        Latin:
 *                            logical order:   <f> <f> <i> <u>
 *                            visual order:    <f> <f> <i> <u>
 *                                        (1)  compose(<f>,   <f>) = <ff>
 *                                        (2)  compose(<ff>,  <i>) = <ffi>
 *                                        (3)  compose(<ffi>, <u>) = 0
 *
 *                        Arabic:
 *                            logical order:   <beh> <lam> <alef>
 *                            visual order:    <alef> <lam> <beh>
 *                                        (1)  compose(<alef>, <lam>)     = <lam-alef>
 *                                        (2)  compose(<lam-alef>, <beh>) = 0
 *
 *                      Note that the composition order in Arabic is (logically)
 *                      backwards.
 *                      This was done to ensure that ligatures are still 'correctly'
 *                      composed if the text is rendered in the direction opposite
 *                      to standard direction, e.g. by directional overrides.
 *                      Furthermore, it is consistent with the joining rules given
 *                      in the Unicode book, Chapter 8.2.  (Although that NSMs are
 *                      not transparent is a clear violation.  Maybe we'll find a
 *                      better, compliant solution later.)
 *
 *    Fonts may use code points outside the Unicode range U+0000..U+10FFFF, e.g. to
 *    store ligatures.  However, only codes in the range 0x0000..0x1fffff are allowed.
 *    This is to ensure that 21 bits are enough to store any code ever processed.
 *    Note that ftx.pl splits the range again: it does allow code points outside
 *    Unicode, but does not dump anything beyond 0x18ffff.  Font designers may use
 *    codes beyond that to store intermediate glyphs or glyph components which will
 *    not be dumped.  However, ftx.pl still generates codes beyond that, namely to
 *    represent intermediate, unassigned glyphs during ligature composition.  These
 *    are in the range 0x190000 to 0x1fffff, still within the 21 bit range.
 */

typedef enum _LiqEncoding {           /* only Unicode font encoding is supported by this library */
    LIQ_ENC_UNICODE
} LiqEncoding;

typedef enum _LiqLanguage {           /* all single-bit constants (currently no more than 32) */
    LIQ_LANG_OTHER = 0x00000000,
    LIQ_LANG_USER1 = 0x00000001,      /* for extensions in user-provided fonts */
    LIQ_LANG_USER2 = 0x00000002,      /* for extensions in user-provided fonts */
    LIQ_LANG_USER3 = 0x00000004,      /* for extensions in user-provided fonts */
    LIQ_LANG_USER4 = 0x00000008,      /* for extensions in user-provided fonts */
    LIQ_LANG_QIP   = 0x00000010,      /* Intern. Phonetic Alphabet (local ISO-649-3 code 'qip') */
    LIQ_LANG_SRP   = 0x00000020,      /* Serbian (glyph variant of CYRILLIC SMALL LETTER BE) */
    LIQ_LANG_MKD   = 0x00000020,      /* same code, because it's the same for Macedonian */
    LIQ_LANG_URD   = 0x00000040,      /* Urdu */
    LIQ_LANG_SND   = 0x00000080,      /* Sindhi */
    LIQ_LANG_FAS   = 0x00000100       /* Persian */
} LiqLanguage;

typedef LiqUInt32 LiqCodePoint;       /* Unicode code point (integer version of LiqChar32) */
typedef LiqUInt32 LiqLanguageSet;     /* set of LiqLanguage as a bit field */
typedef LiqUInt32 LiqGlyph;           /* glyph number */

typedef enum _LiqGlyphFlag {
    /* Glyph (auto-)selection: */
    /* Form (values are synced with LIQ_FLAG_JOIN_*): */
    LIQ_GLYPH_JOIN_MASK  = 0x03,      /* mask for joining */
    LIQ_GLYPH_JOIN_NONE  = 0x00,      /* isolated form */
    LIQ_GLYPH_JOIN_LEFT  = 0x01,      /* left-joining form  (for Arabic, left  = initial) */
    LIQ_GLYPH_JOIN_RIGHT = 0x02,      /* right-joining form (for Arabic, right = final) */
    LIQ_GLYPH_JOIN_BOTH  = 0x03,      /* medial form */

    LIQ_GLYPH_MONO       = 0x04,      /* This glyph is especially suited for mono-spaced mode. */

    LIQ_GLYPH_ALT        = 0x08,      /* Alternative glyph for the immediately precedind glyph,
                                       * to be used to improve automatic kerning.  A sequence
                                       * of alternatives must directly succeed the glyph variant
                                       * selected by the main glyph select algorithm: any
                                       * sequence of glyphs after the best-match selection will
                                       * be considered equal alternatives that may be used if
                                       * more kerning can be achieved by using them.
                                       * Also note that only for the left of two adjacent glyphs,
                                       * alternatives are considered.  If a right glyph has an
                                       * alternative form, ligatures must be used instead.
                                       * Further, kerning to the left is assumed to be the same
                                       * for all alternatives, i.e., an alternative is selected
                                       * without rechecking the left-next kerning again.  In
                                       * short, alternative selection is strictly from
                                       * left to right.
                                       */

    /* Glyph/Character properties: */
    LIQ_GLYPH_FIX_W      = 0x10       /* Glyph has fixed width, no automatic kerning allowed.
                                       * This setting must be consistent for all alternatives
                                       * of a glyphs, otherwise the behaviour is undefined. */
} LiqGlyphFlags;
typedef LiqUInt32 LiqGlyphFlagSet;    /* several flags */

struct _LiqFont {
    int                  api;         /* font API version (current maximum: 1) */
    char const *         name;        /* font name */
    LiqEncoding          encoding;    /* font encoding (see LiqEncoding) */
    LiqX                 width;       /* glyph width in mono-spaced mode */
    LiqY                 height;      /* pixel height */
    LiqX                 bitwidth;    /* maximal glyph width (may be smaller than width) */
    LiqX                 max_overlap; /* maximal allowed negative kerning.  Usually width/2. */
    LiqX                 space_x;     /* default letter spacing in proportional mode */
    LiqBool              mono;        /* LIQ_FALSE for proportional, LIQ_TRUE for mono-spaced */

    LiqCodePoint         (*compose)          (LiqCodePoint, LiqCodePoint, LiqBool);
    LiqGlyph             (*first_glyph)      (LiqCodePoint);
    LiqGlyph             (*next_glyph)       (LiqGlyph);
    LiqX                 (*glyph_x1)         (LiqGlyph);
    LiqX                 (*glyph_x2)         (LiqGlyph);
    LiqColumnData        (*glyph_data)       (LiqGlyph, LiqX);
    LiqLanguageSet       (*glyph_lang_set)   (LiqGlyph);
    LiqColumnData        (*glyph_space_col)  (LiqGlyph);
    LiqX                 (*glyph_space_x)    (LiqGlyph);
    LiqGlyphFlagSet      (*glyph_flags)      (LiqGlyph);
};

#ifdef __cplusplus
}
#endif

#endif /* LIQUID_FONT_H_239843517923124 */
