/* -*- Mode: C++ -*- */

#ifndef LIQ_LIQUID_CXX_10983091409328424
#define LIQ_LIQUID_CXX_10983091409328424

#include <liquid/render.h>

typedef struct _LiqCharInfo LiqCharInfo;

/* NOTE:
 *
 * In modern C++, the following character data types exist:
 *
 *              Minimum Bitsize   May Hold
 *   char       8                 ascii, latin1, utf-8
 *   wchar_t    16                ucs-2
 *   wchar_t    32                ucs-4  (=utf-32)
 *   char16_t   16                utf-16
 *   char32_t   32                utf-32 (=ucs-4)
 *
 * These are all different types and can be overloaded.  Whether
 * wchar_t is 16 or 32 bits depends on the architecture.
 *
 * Now, if this was consistently implemented by the compilers, we
 * could overload set_text with wchar_t and with char16_t and
 * decide by that whether the code is ucs-2 or utf-16.
 * Unfortunately, it is currently not implemented consistently,
 * so an #ifdef orgy would be necessary to do it properly.  That's
 * not yet implemented.
 *
 * OTOH, surrogates are illegal in ucs-2 anyway, so it does not
 * hurt much to assume 16-bit wchar_t might hold utf-16.
 *
 * Still, we need to improve the API so that on a machine that
 * supports all the correct types, you can write:
 *
 *       render.set_text ("blah");
 *       render.set_text (L"blah");  // wchar_t  string
 *       render.set_text (u"blah");  // char16_t string
 *       render.set_text (U"blah");  // char32_t string
 *
 * Currently, what you can write depends on which types are available.
 */

typedef struct _LiqParagraphInfo LiqParagraphInfo;
typedef struct _LiqLineInfo      LiqLineInfo;

typedef struct _LiqBitmapClass {
    void              (*set) (void *, LiqX, LiqColumnData);
    LiqColumnData     (*add) (void *, LiqX, LiqColumnData); /* returns old pixel column */
    LiqColumnData     (*get) (void const *, LiqX);
} LiqBitmapClass;

struct _LiqRender {
#ifndef IN_LIQUID_COMPILING
private:
#endif
    /* These are internal only, but unfortunately need to be exposed if we want to use
     * LiqRender directly.  We could have used a different type and dynamically
     * allocated the internal type, but then C and C++ could not easily be mixed. */

    LiqFont const    *m_font;
    LiqCharInfo      *m_info;
    LiqParagraphInfo *m_paragraph;
    LiqLineInfo      *m_line;
    void             *m_bitmap;
    LiqBitmapClass    m_bitmap_class;
    int               m_info_cnt;
    int               m_paragraph_cnt;
    int               m_line_cnt;
    int               m_width;
    LiqX              m_tracking;
    LIQ_UBITS8        m_bidi      : 1; /* bidi() was done? */
    LIQ_UBITS8        m_auto_kern : 1; /* allow auto-kerning (default: TRUE) */

#ifndef IN_LIQUID_COMPILING
public:
    _LiqRender()
    {
        liq_render_init (this);
    }

    ~_LiqRender()
    {
        liq_render_done (this);
    }

    void set_font (LiqFont const *font)
    {
        liq_render_set_font (this, font);
    }

    void set_paragraph_level (int level)
    {
        liq_render_set_paragraph_level (this, -1, level);
    }

    void set_paragraph_level (LiqParagraph para, int level)
    {
        liq_render_set_paragraph_level (this, para, level);
    }

    void set_tracking (LiqX amount)
    {
        liq_render_set_tracking (this, amount);
    }

    void set_auto_kerning (bool do_auto_kern = true)
    {
        liq_render_set_auto_kerning (this, do_auto_kern);
    }

    void set_text (LiqChar8 const *s)
    {
        liq_render_set_text_utf8 (this, s);
    }

    void set_text (LiqChar8 const *s, size_t n)
    {
        liq_render_set_text_n_utf8 (this, s, n);
    }

    void set_text_latin1 (LiqChar8 const *s)
    {
        liq_render_set_text_latin1 (this, s);
    }

    void set_text_latin1 (LiqChar8 const *s, size_t n)
    {
        liq_render_set_text_n_latin1 (this, s, n);
    }


    /* FIXME: should support wchar_t and char16_t if both are 16 bits. */
    void set_text (LiqChar16 const *s)
    {
        liq_render_set_text_utf16 (this, s);
    }

    void set_text (LiqChar16 const *s, size_t n)
    {
        liq_render_set_text_n_utf16 (this, s, n);
    }


    /* FIXME: should support wchar_t and char32_t if both are 32 bits. */
    void set_text (LiqChar32 const *s)
    {
        liq_render_set_text_utf32 (this, s);
    }

    void set_text (LiqChar32 const *s, size_t n)
    {
        liq_render_set_text_n_utf32 (this, s, n);
    }


    void bidi ()
    {
        liq_render_bidi (this);
    }

    void clear_bidi ()
    {
        liq_render_clear_bidi (this);
    }

    int get_line_count ()
    {
        return liq_render_get_line_count (this);
    }

    int get_paragraph_count ()
    {
        return liq_render_get_paragraph_count (this);
    }

    int get_paragraph_level (LiqParagraph para)
    {
        return liq_render_get_paragraph_level (this, para);
    }

    void bitmap ()
    {
        liq_render_bitmap (this);
    }

    void clear_bitmap ()
    {
        liq_render_clear_bitmap (this);
    }

    LiqParagraph get_line_paragraph (LiqLine line)
    {
        return liq_render_get_line_paragraph (this, line);
    }

    LiqX get_line_width (LiqLine line)
    {
        return liq_render_get_line_width (this, line);
    }

    LiqY get_line_height ()
    {
        return liq_render_get_line_height (this);
    }

    LiqX get_max_width ()
    {
        return liq_render_get_max_width (this);
    }

    LiqColumnData get_column (LiqLine line, LiqX x)
    {
        return liq_render_get_column (this, line, x);
    }
#endif
};

#endif /* LIQ_LIQUID_CXX_10983091409328424 */
