/* -*- Mode: C -*- */

/* COPYING
 * Liquid Rendering Library
 * (c) 2009 by Henrik Theiling
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * SYNOPSIS
 * This is a light-weight library written in portable C (at least
 * that is the hope).  Its focus is on high quality rendering of small
 * bitmap fonts, while still being small.  All necessary algorithms
 * for full Unicode support needed for the rendering are implemented.
 * However, several parts are optional (e.g. UTF-8 and -16 decoding,
 * bidirectional rendering and all the fonts modules) and the linker
 * will drop them if not used, so the memory footprint is minimal.
 */

#ifndef LIQUID_H_1894753974918243210983
#define LIQUID_H_1894753974918243210983

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>

#include <liquid/systype.h>

#ifdef __cplusplus
extern "C" {
#endif

/* RENDERING API
 ***************
 * Text rendering is done in the following steps:
 *
 *    step 1:  liq_render_new()
 *
 *    step 2a: liq_render_set_font()              mandatory
 *    step 2b: liq_render_set_auto_kerning()      (optional: default: LIQ_TRUE)
 *    step 2d: liq_render_set_tracking()          (optional, default: 0)
 *
 *    step 3:  liq_render_set_text_...()          mandatory, will also split
 *                                                text into paragraphs.
 *
 *    step 3o: liq_render_get_paragraph_count()   Paragraphs are enumerated 0..n-1
 *
 *    step 4:  liq_render_set_paragraph_level()   (optional) -1 sets all paragraph levels at once.
 *                                                           The default level is -1, meaning
 *                                                           to automatically derive the level.
 *
 *    step 5:  liq_render_bidi()                  (optional) bidirectional alg. (Hebrew,Arabic,...)
 *
 *    step 5o: liq_render_get_line_count()        get number of lines found in Bidi
 *             liq_render_get_line_paragraph()    find out which paragraph line is in
 *             liq_render_get_paragraph_level()   manually set or derived by bidi()
 *                                                If yet unset, returns -1.  If this is
 *                                                odd, the direction is R2L, otherwise L2R.
 *                                                See liq_level_get_direction.
 *
 *    step 6:  liq_render_bitmap()                the main rendering step
 *
 *    step 6o: liq_render_get_line_width()        using the bitmap
 *             liq_render_get_line_height()
 *             liq_render_get_line_count()
 *             liq_render_get_column()
 *             liq_column_get_pixel()
 *
 *    step 7:  liq_render_delete()
 *
 * The steps ending in 'o' are informational steps where you can query information
 * if you need it.  They do not influence the rendering.
 *
 * In general, invoking a function from a lowel step that changes settings
 * is no error, but will reset the information from higher steps, in
 * particular:
 *
 *    - set_font(), set_paragraph_level(), set_auto_kerning()
 *      => reset text, bidi info and bitmap
 *
 *    - set_text
 *      => resets bidi info and bitmap
 *
 * Because bidi() and bitmap() do not change settings, they may be invoked
 * several times without resetting anything: the duplicate requests are just
 * ignored.
 *
 * The set_text functions also break the text into paragraphs according to the
 * Unicode Bidi algorithm (this step is implemented outside the scope of the
 * Bidi module here).  The preferable paragraph break character is U+2028,
 * but a simple \n (U+000A) also does the job (note that \n is a paragraph
 * break in Unicode, not a line break).
 *
 * liq_render_get_line_count(): Currently the library does no automatic line
 * breaking, only explicit line breaking (defined by Unicode's linebreak.txt,
 * preferable break character is U+2028, but a simple \v (U+000B line
 * tabulation) also does the job).
 *
 * liq_render_set_font() must be used before setting the text, because
 * we need information from the font to decide how much memory to allocate
 * for the text.  There is no default font, you must set one.  This is to
 * avoid a link-dependency: all fonts are in separate modules, so if you
 * don't use them, the linker will drop the whole object file.
 *
 * liq_render_set_paragraph_level() can be used to set the
 * paragraph level for all paragraphs.  By default, it will be determined
 * automatically in the bidi step for each paragraph from its input data,
 * but you might want to change this.  You may either use a numeric number
 * between 0 and 61, even numbers meaning left-to-right text and odd numbers
 * meaning right-to-left, or use LIQ_L2R or LIQ_R2L (which are equivalent
 * to 0 and 1).
 *
 * liq_render_auto_kerning() can be used to switch on/off the automatic kerning,
 * moving two characters closer together just so they do not touch or overlap.
 * This is especially simple for bitmap fonts, and it improves the beauty of the
 * result significantly without much additional effort or manual kerning tables.
 *
 * Running bidi() is optional.  If you render only L2R scripts, you don't need
 * it. You need it for any R2L text (Arabic, Hebrew, ...) even if you do not
 * mix the scripts with L2R scripts.
 * If you do not invoke bidi(), the linker will drop the whole Bidi object
 * file, which saves memory.
 * (set_paragraph_level(), a tiny function, is outside the bidi module, so
 * you do not get a dependency if you invoke it.)
 *
 * Eventually, bitmap() renders the bitmap and you can access it with the
 * get_...() functions.
 * The bitmap is stored as an array of columns from left to right (index 0
 * is left-most), with bit i (with value 2^i) running from top to
 * bottom (0 is top-most) for each line.  The access is encapsulated in small
 * functions/macros for convenience.  Rendering line spacing is thus up
 * to you: just add to the height() the line spacing you desire, that's why
 * the y coordinate is split into two dimensions: line number + y offset
 * for each line.
 * The API is defined in such a way that the internal storage the library
 * uses for the bitmap may be changed without API changes.
 *
 * The algorithm currently does no automatic line breaking.  This is future
 * work.  It does break lines at \n and at paragraph separators in general.
 */

typedef struct _LiqFont LiqFont;       /* the font API is described in liquid-font.h */

typedef enum _LiqDirection {
    LIQ_L2R = 0,                       /* left-to-right text */
    LIQ_R2L = 1                        /* right-to-left text */
} LiqDirection;

typedef unsigned  LiqColumnData;       /* bits of one pixel column */
typedef int       LiqLine;             /* integer used for text line number */
typedef int       LiqParagraph;        /* integer used for paragraph number */
typedef int       LiqX;                /* integer used for pixel x coordinate (must be signed) */
typedef int       LiqY;                /* integer used for pixel y coordinate (must be signed) */

typedef struct _LiqRender LiqRender;   /* the main renderer */

typedef void * (*LiqCalloc) (size_t, size_t);
                                       /* memory allocation: calloc (default: LIQ_CALLOC) */
typedef void   (*LiqFree)   (void *);  /* memory allocation: free   (default: LIQ_FREE) */

extern void           liq_set_allocator                  (LiqCalloc, LiqFree);
                                                         
extern LiqRender *    liq_render_new                     (void);
extern void           liq_render_init                    (LiqRender *);
extern void           liq_render_done                    (LiqRender *);
extern void           liq_render_delete                  (LiqRender *);
                                                         
extern void           liq_render_set_font                (LiqRender *, LiqFont const *);
extern void           liq_render_set_tracking            (LiqRender *, LiqX); /* letter-spacing */
extern void           liq_render_set_auto_kerning        (LiqRender *, LiqBool);

extern void           liq_render_set_text_utf8           (LiqRender *, LiqChar8  const *);
extern void           liq_render_set_text_n_utf8         (LiqRender *, LiqChar8  const *, size_t);
extern void           liq_render_set_text_latin1         (LiqRender *, LiqChar8  const *);
extern void           liq_render_set_text_n_latin1       (LiqRender *, LiqChar8  const *, size_t);
extern void           liq_render_set_text_utf16          (LiqRender *, LiqChar16 const *);
extern void           liq_render_set_text_n_utf16        (LiqRender *, LiqChar16 const *, size_t);
extern void           liq_render_set_text_utf32          (LiqRender *, LiqChar32 const *);
extern void           liq_render_set_text_n_utf32        (LiqRender *, LiqChar32 const *, size_t);
extern void           liq_render_clear_text              (LiqRender *);

extern int            liq_render_get_paragraph_count     (LiqRender const *);
extern void           liq_render_set_paragraph_level     (LiqRender *, LiqParagraph, int);

extern void           liq_render_bidi                    (LiqRender *);
extern void           liq_render_clear_bidi              (LiqRender *);
extern int            liq_render_get_line_count          (LiqRender const *);
extern int            liq_render_get_paragraph_level     (LiqRender const *, LiqParagraph);

extern void           liq_render_bitmap                  (LiqRender *);
extern void           liq_render_clear_bitmap            (LiqRender *);
extern LiqX           liq_render_get_line_width          (LiqRender const *, LiqLine);
extern LiqParagraph   liq_render_get_line_paragraph      (LiqRender const *, LiqLine);
extern LiqY           liq_render_get_line_height         (LiqRender const *);
extern LiqX           liq_render_get_max_width           (LiqRender const *);
extern LiqColumnData  liq_render_get_column              (LiqRender const *, LiqLine, LiqX);

#define liq_column_get_pixel(column,y) (!!((column) & (1U << (y))))

#define liq_level_get_direction(level) (((level) & 1) ? LIQ_R2L : LIQ_L2R)

/* Fonts: */
extern LiqFont const *const liq_font_liqminus;
extern LiqFont const *const liq_font_liqminusmono;
extern LiqFont const *const liq_font_liqsting;
extern LiqFont const *const liq_font_liqstingmono;
extern LiqFont const *const liq_font_liqmean;
extern LiqFont const *const liq_font_liqmeanmono;
extern LiqFont const *const liq_font_liqwee;
extern LiqFont const *const liq_font_liqweemono;

#ifdef __cplusplus
} /* extern "C" */
#endif

/* This library uses precompiled tables derived from official Unicode
 * data tables plus private extensions for the private use areas.
 * The particular files that were used are:
 *
 *     Official Unicode               Private Extensions
 *     ---------------------------------------------------
 *     unicodedata.txt                unidata-priv.dat
 *     bidimirroring.txt              bidimirr-priv.dat
 *     linebreak.txt                  linebreak-priv.dat
 *     derivedcoreproperties.txt
 *  ([ scripts.txt                    scripts-priv.dat  ])
 *   [ blocks.txt                    (blocks-priv.data) ]
 *
 * Those in () are not yet implemented or used.
 * Those in [] are not used for this library, but for other purposes,
 * e.g. pdf generation for the Liquid font family.
 *
 * - arabicshaping.txt is currently not used, because we use whatever
 *   joining types the font defines.  The algorithm is what Unicode
 *   defines, however.
 *
 * - derivedcoreproperties.txt is used to find the Default Ignorable
 *   code points.
 */

#endif /* LIQUID_H_1894753974918243210983 */
