/* -*- Mode: C++ -*- */

/* COPYING
 * Liquid Rendering Library
 * (c) 2009 by Henrik Theiling
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "internal.h"

/* ******************************************************************************** */

static void *liq_calloc_default (size_t n, size_t s)
{
    return LIQ_CALLOC(n, s);
}

static void liq_free_default (void *x)
{
    LIQ_FREE(x);
}

/* ******************************************************************************** */

LiqCharInfo const liq_char_info_0= LIQ_CHAR_INFO_0;
LiqCalloc liq_calloc= liq_calloc_default;
LiqFree   liq_free=   liq_free_default;

/* ******************************************************************************** */
/* BIDI category NSM: generated list for bsearch(): */

typedef struct _NSMRange {
    LIQ_UBITS32 start : 32; /* keep this word aligned to make access faster */
    LIQ_UBITS32 end   : 24; /* minimum 21, because we don't know signedness */
    LIQ_UBITS8  nsm   : 8;  /* NSM or B all others are handled in bidi.c */
} NSMRange;

static NSMRange const nsm_range[]= {
#include "autogen/bidi1.inc"
};

static size_t const nsm_range_cnt= sizeof(nsm_range) / sizeof(nsm_range[0]);

static int nsm_cmp (void const *a, void const *b)
{
    LiqChar32 key=             *(LiqChar32 const *)a;
    NSMRange const *arr_elem=  (NSMRange const *)b;
    return key < (LiqChar32)arr_elem->start ?
               -1
           : key > (LiqChar32)arr_elem->end ?
               +1
           :   0;
}

static LiqBidiType liq_get_bidi_base_type (LiqChar32 c)
{
    NSMRange *range=
        (NSMRange*)bsearch (
            &c,
            nsm_range,
            nsm_range_cnt,
            sizeof(nsm_range[0]),
            nsm_cmp);
    if (range)
        return range->nsm;
    return LIQ_BIDI_ON;
}

/* ******************************************************************************** */
/* Default ignorable */

typedef struct _IgnoreRange {
    LiqChar32 start; /* keep this word aligned to make access faster */
    LiqChar32 end;   /* minimum 21, because we don't know signedness */
} IgnoreRange;

static IgnoreRange const ignore_range[]= {
#include "autogen/ignore.inc"
};

static size_t const ignore_range_cnt= sizeof(ignore_range) / sizeof(ignore_range[0]);

static int ignore_cmp (void const *a, void const *b)
{
    LiqChar32 key=               *(LiqChar32 const *)a;
    IgnoreRange const *arr_elem=  (IgnoreRange const *)b;
    return key < (LiqChar32)arr_elem->start ?
               -1
           : key > (LiqChar32)arr_elem->end ?
               +1
           :   0;
}

static LiqBool liq_get_important (LiqChar32 c)
{
    IgnoreRange *range=
        (IgnoreRange*)bsearch (
            &c,
            ignore_range,
            ignore_range_cnt,
            sizeof(ignore_range[0]),
            ignore_cmp);
    if (range)
        return LIQ_FALSE;
    return LIQ_TRUE;
}

/* ******************************************************************************** */

void liq_set_allocator (LiqCalloc c, LiqFree f)
{
    liq_calloc= c;
    liq_free=   f;
}

void liq_render_init (LiqRender *self)
{
    memset (self, 0, sizeof(*self));
    self->m_auto_kern=  LIQ_TRUE;
}

void liq_render_clear_text (LiqRender *self)
{
    liq_render_clear_bitmap (self);

    TFREE(self->m_line);
    self->m_line_cnt= 0;

    TFREE(self->m_paragraph);
    self->m_paragraph_cnt= 0;

    TFREE(self->m_info);
    self->m_info_cnt= 0;

    self->m_bidi= LIQ_FALSE;
}

void liq_render_done (LiqRender *self)
{
    liq_render_clear_text(self);
}

LiqRender *liq_render_new (void)
{
    LiqRender *self= TCALLOC(LiqRender,1);
    liq_render_init (self);
    return self;
}

void liq_render_delete (LiqRender *self)
{
    if (self) {
        liq_render_done (self);
        TFREE (self);
    }
}

void liq_render_clear_bidi (LiqRender *self)
{
    liq_render_clear_bitmap (self);
    if (self->m_bidi)
        liq_render_clear_text(self);
}

/* global settings: */
void liq_render_set_font (LiqRender *self, LiqFont const *font)
{
    liq_render_clear_text (self);
    self->m_font= font;
    assert (font->api == 1);
}

void liq_render_set_paragraph_level(LiqRender *self, LiqParagraph para, int level)
{
    assert (level >= 0 && level <= 61);
    assert (self->m_paragraph);
    if (para < 0)
        for (para=0; para < self->m_paragraph_cnt; ++para)
            self->m_paragraph[para].level= level;
    else
        self->m_paragraph[para].level= level;
    liq_render_clear_bidi (self);
}

int liq_render_get_paragraph_level(LiqRender const *self, LiqParagraph para)
{
    assert (self->m_paragraph);
    assert (para >= 0);
    assert (para <= self->m_paragraph_cnt);
    return self->m_paragraph[para].level;
}

void liq_render_set_auto_kerning (LiqRender *self, LiqBool value)
{
    liq_render_clear_text (self);
    self->m_auto_kern= !!value;
}

void liq_render_set_tracking (LiqRender *self, LiqX amount)
{
    liq_render_clear_text (self);
    self->m_tracking= amount;
}

/* bitmap access: */
int liq_render_get_line_count (LiqRender const *self)
{
    assert (self->m_line && "expected liq_render_bitmap() first");
    return self->m_line_cnt;
}

int liq_render_get_paragraph_count (LiqRender const *self)
{
    assert (self->m_paragraph && "expected liq_render_bitmap() first");
    return self->m_paragraph_cnt;
}

LiqParagraph liq_render_get_line_paragraph (LiqRender const *self, LiqLine l)
{
    assert (self->m_line && "expected liq_render_bitmap() first");
    return self->m_line[l].paragraph - self->m_paragraph;
}

LiqX liq_render_get_line_width (LiqRender const *self, LiqLine l)
{
    assert (self->m_line && "expected liq_render_bitmap() first");
    assert (l >= 0 && l < liq_render_get_line_count(self));
    return self->m_line[l].pixel_width;
}

LiqX liq_render_get_max_width (LiqRender const *self)
{
    int i;
    LiqX max_w= 0;
    for (i=0; i < self->m_line_cnt; i++) {
        LiqX w= liq_render_get_line_width (self, i);
        if (w > max_w) max_w= w;
    }
    return max_w;
}

LiqY liq_render_get_line_height (LiqRender const *self)
{
    assert (self->m_font && self->m_bitmap && "expected liq_render_bitmap() first");
    return self->m_font->height;
}

/* set text */
void liq_render_alloc_info (LiqRender *self, size_t n)
{
    liq_render_clear_bidi (self);
    self->m_info= TCALLOC (LiqCharInfo, n);
    self->m_info_cnt= n;
}

/* ********************************************************************** */

static LiqBreakType liq_break_type_simple (LiqChar32 c)
{
#define TYPE(X) LIQ_BREAK_##X
#include "autogen/lbsimple.inc"
#undef  TYPE
    return LIQ_BREAK_XX;
}

void liq_set_info (LiqCharInfo *i, LiqChar32 c)
{
    *i= liq_char_info_0;
    i->code=      c;
    i->bidi=      liq_get_bidi_base_type (c);
    i->important= liq_get_important (c);

    if (liq_break_type_simple (c) == LIQ_BREAK_BK)
        i->line_break= LIQ_BREAK_AND_DROP;
}

void liq_seq_reverse (LiqCharInfo *first, LiqCharInfo *last)
{
    assert (first != NULL);
    assert (last  != NULL);

    while (first < last) {
        LiqCharInfo help= *first;
        *first= *last;
        *last= help;
        first++;
        last--;
    }
}

static void liq_paragraph_init (
    LiqParagraphInfo *para, LiqCharInfo *first, LiqCharInfo *last)
{
    assert (para  != NULL);
    assert (first != NULL);
    assert (last  != NULL);

    para->first= first;
    para->last=  last;
    para->level= -1;
}

void liq_split_paragraphs (LiqRender *self)
{
    LiqCharInfo *const first= self->m_info;
    LiqCharInfo *const last=  self->m_info + self->m_info_cnt - 1;

    if (self->m_paragraph) /* already done? */
        return;

    /* pass 1 (self->m_paragraph == NULL): count
     * pass 2 (self->m_paragraph != NULL): fill in */
    forever {
        int i=0;
        LiqCharInfo *begin= first;
        LiqCharInfo *cur;
        for (cur= first; cur <= last; ++cur) {
            if (cur->bidi == LIQ_BIDI_B) {
                if (begin < cur) {
                    if (self->m_paragraph)
                        liq_paragraph_init (self->m_paragraph+i, begin, cur-1);
                    i++;
                }
                begin= cur+1;
            }
        }
        if (begin <= last) {
            if (self->m_paragraph)
                liq_paragraph_init (self->m_paragraph + i, begin, last);
            i++;
        }

        if (self->m_paragraph) {
            assert (i == self->m_paragraph_cnt);
            break; /*!*/
        }

        self->m_paragraph= TCALLOC(LiqParagraphInfo, i);
        self->m_paragraph_cnt= i;
    }
}

void liq_render_finalise_text  (LiqRender *self)
{
    /* P1 (works without bidi classes) */
    liq_split_paragraphs (self);
}

static void liq_line_init (
    LiqLineInfo *line, LiqParagraphInfo *para, LiqCharInfo *first, LiqCharInfo *last)
{
    assert (line  != NULL);
    assert (para  != NULL);
    assert (first != NULL);
    assert (last  != NULL);

    line->paragraph= para;
    line->first=     first;
    line->last=      last;
}

/*
 * Paragraph breaks (Bidi class B):
 *     000A LINE FEED (LF)
 *     000D CARRIAGE RETURN (CR)
 *     001C INFORMATION SEPARATOR FOUR
 *     001D INFORMATION SEPARATOR THREE
 *     001E INFORMATION SEPARATOR TWO
 *     0085 NEXT LINE (NEL)
 *     2029 PARAGRAPH SEPARATOR
 *
 * Explicit line breaks (paragraph breaks in parens):
 *   ( 000A LINE FEED (LF)       )
 *     000B LINE TABULATION
 *     000C FORM FEED (FF)
 *   ( 000D CARRIAGE RETURN (CR) )
 *   ( 0085 NEXT LINE (NEL)      )
 *     2028 LINE SEPARATOR
 *   ( 2029 PARAGRAPH SEPARATOR  )
 *
 * Recognition of the explicit line breaks without the paragraph breaks
 * are generated by linebreak.pl (and map2c.pl).
 */
void liq_split_lines (LiqRender *self)
{
    assert (self != NULL);

    if (self->m_line) /* already done? */
        return;

    /* pass 1 (self->m_line == NULL): count
     * pass 2 (self->m_line != NULL): fill in */
    forever {
        int i= 0;
        int o;
        for (o=0; o < self->m_paragraph_cnt; o++) {
            LiqParagraphInfo *para=  self->m_paragraph + o;
            LiqCharInfo  *begin= para->first;
            LiqCharInfo  *cur;
            for (cur= para->first; cur <= para->last; ++cur) {
                if (cur->line_break != LIQ_BREAK_NONE) {
                    LiqCharInfo *last= cur - (cur->line_break & LIQ_BREAK_BEFORE ? 1 : 0);
                    if (begin <= last) {
                        if (self->m_line)
                            liq_line_init (self->m_line + i, para, begin, last);
                        i++;
                    }
                    begin= cur + (cur->line_break & LIQ_BREAK_AFTER ? 1 : 0);
                }
            }
            if (begin <= para->last) {
                if (self->m_line)
                    liq_line_init (self->m_line + i, para, begin, para->last);
                i++;
            }
        }

        if (self->m_line) {
            assert (i == self->m_line_cnt);
            break; /*!*/
        }

        self->m_line= TCALLOC(LiqLineInfo, i);
        self->m_line_cnt= i;
    }
}

void liq_foreach_run (
    LiqRender    *self,
    LiqParagraphInfo *para,
    void (*step) (
        LiqRender *, LiqParagraphInfo *,
        int, LiqBidiType, LiqBidiType, LiqCharInfo *, LiqCharInfo *))
{
    LiqCharInfo *begin= para->first;
    LiqCharInfo *cur;
    int run_level= para->level;
    LiqBidiType sor= level2bidi(para->level);
    for (cur= para->first; cur <= para->last; ++cur) {
        if (cur->bidi == LIQ_BIDI_BN)
            continue;

        if (cur->level != run_level) {
            int eor= level2bidi (max_int (run_level, cur->level));
            step (self, para, run_level, sor, eor, begin, cur-1);
            sor=       eor;
            begin=     cur;
            run_level= cur->level;
        }
    }

    if (begin <= para->last) {
        int eor= level2bidi(max_int (run_level, para->level));
        step (self, para, run_level, sor, eor, begin, para->last);
    }
}

/* ********************************************************************** */

LiqColumnData liq_render_get_column (LiqRender const *self, LiqLine l, LiqX x)
{
    assert (self && self->m_bitmap && "expected liq_render_bitmap() first");
    assert (l >= 0 && l < liq_render_get_line_count(self));
    assert (x >= 0 && x < liq_render_get_line_width(self,l));
    return self->m_bitmap_class.get (self->m_bitmap, self->m_line[l].pixel_start + x);
}
