/* -*- Mode: C++ -*- */

/* COPYING
 * Liquid Rendering Library
 * (c) 2009 by Henrik Theiling
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "internal.h"

#define LIQ_MAX_HISTORY 32
#define VARIANT_AUTO    -1

static INLINE LiqX glyph_width (LiqFont const *font, LiqGlyph glyph)
{
    return font->glyph_x2(glyph) - font->glyph_x1(glyph) + 1;
}

static INLINE LiqColumnData glyph_get_left_column (
    LiqFont const *font,
    LiqGlyph glyph,
    LiqBool  mirror,
    LiqX     x)
{
    LiqX x1= font->glyph_x1 (glyph);
    LiqX x2= font->glyph_x2 (glyph);

    return
        x < 0 ?
            0
        : mirror ?
            ((x2 - x < x1) ? 0 : font->glyph_data (glyph, x2 - x))
        :   ((x1 + x > x2) ? 0 : font->glyph_data (glyph, x1 + x));
}

static INLINE LiqColumnData glyph_get_right_column (
    LiqFont const *font,
    LiqGlyph glyph,
    LiqBool  mirror,
    LiqX     x)
{
    return glyph_get_left_column (font, glyph, !mirror, x);
}

static INLINE LiqBool bits_touching (LiqColumnData a, LiqColumnData b, LiqColumnData x)
{
    a &= ~x;
    b &= ~x;
    return !!(a & (b | (b << 1) | (b >> 1)));
}

static INLINE LiqColumnData get_history (
    LiqColumnData const *history, int i)
{
    unless (i >= 0)
        return 0;

    unless (i < LIQ_MAX_HISTORY)
        return ~(LiqColumnData)0;

    return history[i];
}

/* Auto-kerning is quite straight-forward for Latin, but a bit harder for Arabic.
 * For non-connected scripts (i.e., space_col == 0), assuming zero spaces between
 * the glyphs, we do not allow overlap of two pixels, nor touching the next higher
 * or lower pixel.  In the following examples, \ is a pixel of the left glyph,
 * / a pixel of the right glyph, and X is an overlapping pixel from both glyphs:
 *
 *                overlap,     touching,
 *     allowed:   forbidden:   forbidden:
 *     \\..       \\..         \\.
 *     ..//       \X//         .//
 *     \\..       .///         \\.
 *
 * We then add the required amount of space_x.  For larger fonts, we would want to
 * check that there is no pixel in a larger radius than just 1 pixel, but we
 * expect very small fonts here, so we're ok.
 *
 * For Arabic and other connected scripts, the pixels set in the space_col are
 * special: they are ignored for normal kerning rules as shown above.  However,
 * if any space_col pixels overlap, then maximally one further pixel of kerning
 * is done.  If space_x is 0, no further kerning is done.  This way, you can
 * set one pixel in both adjacent glyphs defining a joint point for maximum
 * kerning.
 */
static INLINE int liq_auto_kern (
    LiqFont const *font,
    LiqX space_x,
    LiqColumnData space_col,
    LiqColumnData const *history,
    LiqX prev_kern,
    LiqGlyph left,
    LiqBool  left_mirror,
    LiqGlyph right,
    LiqBool  right_mirror)
{
    LiqX k;
    LiqX join_x=  space_x ? 2 : 1;
    LiqX left_w=  glyph_width (font, left);
    LiqX right_w= glyph_width (font, right);
    LiqX max_w=   min_int (  /* never kern more than the character width */
                      font->max_overlap,
                      min_int (left_w, right_w) /* + 1 do not allow complete overlap */
                  );

    /* if there is a joining at the outer pixel, allow one more pixel for kerning */
    if (glyph_get_left_column  (font, right, right_mirror, 0) == space_col &&
        glyph_get_right_column (font, left,  left_mirror,  0) == space_col)
    {
        max_w++;
    }

    for (k=0; k < max_w; k++) {
        LiqX i;
        for (i=0; i <= k; i++) {
            LiqColumnData rbits= glyph_get_left_column  (font, right, right_mirror, i);
            LiqColumnData lbits= glyph_get_right_column (font, left,  left_mirror,  k - i);
            if (bits_touching (
                    rbits,
                    (lbits | get_history (history, k - i - left_w - prev_kern)),
                    space_col))
            {
                return -k;
            }

            if (space_col &&
                (space_col & lbits & glyph_get_left_column (font, right,right_mirror, i-join_x)))
            {
                return -k;
            }
        }
    }

    return -max_w;
}

static INLINE void history_add (
    LiqFont const *font,
    LiqColumnData *history,
    LiqX prev_kern,
    LiqGlyph glyph,
    LiqBool  mirror)
{
    LiqX w=     glyph_width (font, glyph);
    LiqX shift= w + prev_kern;
    int i;

    assert (w     < LIQ_MAX_HISTORY);
    assert (shift < LIQ_MAX_HISTORY);

    for (i=font->max_overlap-1; i >= shift; i--)
        history[i]= history[i - shift];

    for (i=0; i < shift; i++)
        history[i]= 0;

    for (i=0; i < w; i++)
        history[i]|= glyph_get_right_column (font, glyph, mirror, i);
}

static LiqGlyph first_glyph (LiqFont const *font, LiqChar32 c)
{
    LiqGlyph glyph;

    if ((glyph= font->first_glyph (c)) != 0)
        return glyph;

    if ((glyph= font->first_glyph (UNICODE_MISSING_GLYPH)) != 0)
        return glyph;

    if ((glyph= font->first_glyph (UNICODE_REPLACEMENT_CHARACTER)) != 0)
        return glyph;

    if ((glyph= font->first_glyph ('?')) != 0)
        return glyph;

    return 0;
}

static INLINE LiqX kern_clamp (LiqX kern)
{
    if (kern > LIQ_KERN_MAX) return LIQ_KERN_MAX;
    if (kern < LIQ_KERN_MIN) return LIQ_KERN_MIN;
    return kern;
}

/*
 * Glyph selection works in several steps:
 *   TODO:
 *   Step 1: (optional, directly done upon setting the text):
 *           Resolve decompositions and compatibility decompositions (incl.
 *           presentation forms) if the visual formatting information can
 *           be preserved, and if the code point is not in the font.
 *           - presentation forms:
 *                 [FB53 final beh]
 *             --> [0628 bah] + LIQ_FLAG_DO_JOIN_BEFORE  (<final> can be preserved)
 *             E.g. <circle> cannot currently be preserved (among others)
 *
 *           - ligature will be decomposed, too, if the font does not
 *             provide them (i.e, <compat> is preserved by being ignored...)
 *
 *           - whether implicit joining behaviour will override ZWJ/ZWNJ
 *             or not has to be decided.  (e.g. ZWNJ + <final beh>?)  It
 *             is felt that the implicit information should be of higher
 *             priority, i.e., ZWJ/ZWNJ should be ignored.
 *
 *           It should be noted that for those code points present in the
 *           font, the joining behaviour is not known: e.g. only the basic
 *           Arabic characters are properly joined automatically.  So
 *           a decomposed ligature might be automatically joined while
 *           a non-decomposed liguture might not be.  This is inconsistent.
 *           Should all decompositions implicitly default to ZWNJs on their
 *           edges (unless, of course, they are explicitly <medial>, <final>
 *           or <initial>) just to enforce consistency?  This would mean
 *           that, assuming all ligatures exist in the font,
 *                  <f>+<fi>     -->  <f>+<fi>
 *           while  <f>+<f>+<i>  -->  <ffi>
 *           Is that what we want?
 *           What about base + combining (this is not <compat>, so easily
 *           distinguishable):
 *                  <a dot below> <acute>  -->  <a> <dot below> <acute>
 *           -->    <a acute> <dot below>
 *           Probably, if possible.
 *
 *   Step 2: [bidi module]
 *           Final changes to code points:
 *           - mirrored characters if available possible
 *           This means we need a codepoint for all mirrored chars
 *           we support.
 *
 *   TODO:
 *   Step 3: Selection of an appropriate codepoint.  This should:
 *           - combine ligatures
 *           - characters + combining characters:
 *           This means we need a codepoint for all combined chars,
 *           including all ligatures we support.
 *
 *   Step 4: Auto-joining: with the correct codepoint selected,
 *           we can check which chars can joint left and right.
 *
 *   Step 5: Selection of the exact glyph to use:
 *           - manually selected variant (variation selector)
 *           - initial/left/medial/right forms
 *             - which joining is required must already be marked
 *               in the flags.  Since auto-joining is only done
 *               for scripts that require bidi() (Arabic, Syriac, N'ko),
 *               we require bidi() to set the flags.
 *           - proportional vs. monospaced form
 *
 *   Step 6: Kerning: try to reduce the x-space between glyphs.
 *           This may select alternative glyphs again.  This
 *           also sets the amount of kerning, relative to the
 *           normal width of a glyph, namely:
 *               x2 - x1 + 1
 *           So this also resolves the normal x_spacing.
 *
 * This codes implements Steps 2 and 3.
 */
static void liq_kern_pair (
    LiqRender *const self,
    LiqFont const *const font,
    LiqColumnData *const history,
    LiqX *const prev_kern,
    LiqCharInfo *const cur,
    LiqCharInfo *const next)
{
    LiqX kern= 0;
    LiqX const space_x=      font->glyph_space_x   (cur->code);
    LiqColumnData space_col= font->glyph_space_col (cur->code);

    assert (cur->is_glyph);
    assert (next->is_glyph);

    /* find minimal kerning, if possible (and if allowed) */
    if (self->m_auto_kern &&
        !(font->glyph_flags (cur->code)  & LIQ_GLYPH_FIX_W) &&
        !(font->glyph_flags (next->code) & LIQ_GLYPH_FIX_W))
    {
        LiqGlyph glyph= cur->code;
        do {
            int alt_kern=
                liq_auto_kern (font, space_x, space_col, history, *prev_kern,
                    glyph,      cur->mirror,
                    next->code, next->mirror);

            if (alt_kern < kern) {
                kern= alt_kern;
                cur->code= glyph;
            }

            glyph= font->next_glyph(glyph);
        } while (glyph && (font->glyph_flags (glyph) & LIQ_GLYPH_ALT));

        /* insert old char into history: */
        history_add (
            font, history, *prev_kern,
            cur->code, cur->mirror);
    }

    /* total kerning adds minimum of space_x: */
    kern+= space_x + self->m_tracking;

    /* clamp and set: */
    kern= kern_clamp (
              max_int (
                  kern,
                  -glyph_width(font,cur->code) /* never step back */
              )
          );
    cur->kern=  kern;
    *prev_kern= kern;

#if 0
    fprintf (stderr, "clamp(kern + space)= %d\n", kern);
    fprintf (stderr, "history:\n");
    {
        int x;
        int y;
        for (y=0; y < font->height; y++) {
            fprintf (stderr, "    ");
            for (x=LIQ_MAX_HISTORY-1; x >= 0; x--) {
                fprintf (stderr, "%c",
                    (history[x] & (1 << y)) ? '#' : '.');
            }
            fprintf (stderr, "\n");
        }
    }
#endif
}

static void liq_run_kerning (
    LiqRender    *self,
    LiqParagraphInfo *para,
    int           run_level,
    LiqBidiType   sor,
    LiqBidiType   eor,
    LiqCharInfo  *first,
    LiqCharInfo  *last)
{
    LiqCharInfo *cur;
    LiqCharInfo *next;                     
    LiqFont const *const font= self->m_font;
    LiqColumnData history[LIQ_MAX_HISTORY];
    LiqX prev_kern;

    (void)para;
    (void)sor;
    (void)eor;

    assert (font != NULL);
    assert (font->max_overlap < LIQ_MAX_HISTORY);

    memset (history, ~0, sizeof(history)); /* fill with 1s */
    prev_kern= 0;
    cur= NULL;
    if (is_even (run_level)) {
        /* left-to-right: */
        for (next= first; next <= last; ++next) {
            assert (next->kern == 0); /* must not be set this yet */
            if (next->code) {         /* skip invalid glyphs */
                if (cur != NULL)
                    liq_kern_pair (self, font, history, &prev_kern, cur, next);

                cur= next;
            }
        }
    }
    else {
        /* right-to-left: */
        /* We render NSMs on the left side in R2L mode, so this is just
         * like the code above, only in opposite order.  Stray NSMs the
         * have a glyph are treated just like any other glyph here.
         */
        for (next= last; next >= first; --next) {
            assert (next->kern == 0); /* must not be set this yet */
            if (next->code) {         /* skip invalid glyphs */
                if (cur != NULL)
                    liq_kern_pair (self, font, history, &prev_kern, cur, next);

                cur= next;
            }
        }
    }

    if (cur)
        cur->kern= font->glyph_space_x(cur->code);
}

static LiqBool liq_can_join (LiqFont const *font, LiqChar32 code, LiqJoining joining)
{
    LiqGlyph glyph;
    for (glyph= font->first_glyph (code); glyph; glyph= font->next_glyph(glyph))
        if ((font->glyph_flags (glyph) & LIQ_GLYPH_JOIN_MASK) == joining)
            return LIQ_TRUE;
    return LIQ_FALSE;
}

static void liq_pair_select_ligature (LiqFont const *font, LiqCharInfo **left, LiqCharInfo *right)
{
    if (right->important && right->opaque) {
        if (*left &&
            !((*left)->joining & LIQ_NO_JOIN_RIGHT) &&
            !(right->joining   & LIQ_NO_JOIN_LEFT))
        {
            LiqBool force=    ((*left)->joining & LIQ_JOIN_RIGHT) &&
                              (right->joining   & LIQ_JOIN_LEFT);
            LiqCodePoint lig= font->compose ((*left)->code, right->code, force);
            if (lig) {
                LiqJoining lig_joining= ((*left)->joining & LIQ_JOIN_LEFT) |
                                        (right->joining   & LIQ_JOIN_RIGHT);

                if (liq_can_join (font, lig, lig_joining)) {
                    (*left)->code=    lig;
                    (*left)->joining= lig_joining;
                    *right= liq_char_info_0;
                    return; /* do not reset *left */
                }
            }
        }
        *left= right;
    }
}

/* Selecting ligatures might delete code points from the run.
 * We will overwrite them with empty cells.  liq_para_select_ligatures
 * could remove them from the paragraph, but we won't, because we
 * want to test how we handle them later (they should be ignored). */
static void liq_run_select_ligatures (
    LiqRender *self,
    LiqParagraphInfo *para,
    int level,
    LiqBidiType sor,
    LiqBidiType eor,
    LiqCharInfo *first,
    LiqCharInfo *last)
{
    LiqFont const *const font= self->m_font;
    LiqCharInfo *cur;
    LiqCharInfo *pred= NULL;

    (void)sor;
    (void)eor;
    (void)para;

    if (is_odd(level))
        for (cur= last; cur >= first; --cur)
            liq_pair_select_ligature (font, &pred, cur);
    else
        for (cur= first; cur <= last; ++cur)
            liq_pair_select_ligature (font, &pred, cur);
}

/*
 * Selecting ligatures works run-wise, because we need visual order.
 */
void liq_para_select_ligatures (LiqRender *self, LiqParagraphInfo *para)
{
    /* compose ligatures: */
    liq_foreach_run (self, para, &liq_run_select_ligatures);

    /* we don't remove neutrals: the following algorithms must
     * handle BN and completely ignore them. */
}


#if 0
/* TESTING */
void liq_para_remove_bn (LiqRender *self, LiqParagraphInfo *para)
{
    LiqCharInfo *src= para->first;
    LiqCharInfo *dst= para->first;
    while (src <= para->last) {
        unless (src->bidi == LIQ_BIDI_BN) {
            *dst= *src;
            dst++;
        }
        src++;
    }

    src= dst;
    while (dst <= para->last)
        *dst++= liq_char_info_0;

    para->last= src;
}
#endif


/* Final glyph selection & kerning
 **********************************
 * Step 1: Glyph Selection:
 *
 * Selecting glyphs works in logical order, because line-breaking
 * comes later.  Therefore, kerning is more complicated than one
 * might expect.  We must do it run-wise, each run in its own
 * direction.
 *
 * Step 2: Kerning:
 *
 * The kerning between two adjacent glyphs is set at the first
 * glyphs in graphical direction.  The spacing is the minimum
 * of the two glyph spacings.
 *
 * Further, because we do not know the adjacent glyphs at
 * run-borders yet:
 *
 *    - at run-borders, there is no kerning
 *
 *    - the last (in graphical order) glyph of a paragraph has
 *      no spacing.
 *
 *    - the last (in graphical order) glyph of a run uses the
 *      minimum of the font's space_x and the glyph's.
 *
 * Later, we *could* improve kerning, but I think it makes no
 * sense at run-borders anyway.
 */
void liq_para_select_glyphs (LiqRender *self, LiqParagraphInfo *para)
{
    LiqCharInfo *cur;
    LiqFont const *const font= self->m_font;
    LiqCharInfo *const first=  para->first;
    LiqCharInfo *const last=   para->last;

    if (first->is_glyph) /* glyph selection already done? */
        return;

    /* glyph selection */

    /* This replaces all 'code' fields in first..last by glyph IDs.  I.e., the
     * original code point values are lost!  The fact is marked with the bit
     * 'IS_GLYPH' in each char info in order to make assertions possible. */
    cur= first;
    while (cur <= last) {
        LiqCharInfo *nsm= cur+1;  

        if (!cur->important || cur->bidi == LIQ_BIDI_BN) {
            cur->code= 0; /* glyph_id 0 is invalid, so this will be skipped */
        }
        else {
            LiqGlyph glyph;
            int forced_variant= VARIANT_AUTO;
            assert (!cur->is_glyph);

            /* parse NSMs */
            while (nsm <= last && nsm->orig_bidi == LIQ_BIDI_NSM) {
                LiqChar32 c= nsm->code;
                nsm->code= 0; /* skip glyph by default */
                nsm->is_glyph= LIQ_TRUE;

                if (c >= UNICODE_VARIATION_SELECTOR_1 &&
                    c <= UNICODE_VARIATION_SELECTOR_16)
                {
                    forced_variant= c - UNICODE_VARIATION_SELECTOR_1;
                }
                else
                if (nsm->code >= UNICODE_VARIATION_SELECTOR_17 &&
                    nsm->code <= UNICODE_VARIATION_SELECTOR_256)
                {
                    forced_variant= c - UNICODE_VARIATION_SELECTOR_17 + 16;
                }
                else {
                    if ((glyph= first_glyph (font, c)) != 0)
                        nsm->code= glyph;
                }

                nsm++;
            }

            /* Select glyph: */
            if ((glyph= first_glyph (font, cur->code)) != 0) {
                cur->code= glyph; /* by default, select first glyph */
                if (forced_variant != VARIANT_AUTO) {
                    int variant= 0;
                    for (; glyph; glyph= font->next_glyph(glyph)) {
                        if (variant == forced_variant) {
                            cur->code= glyph;
                            break;
                        }
                        variant++;
                    }
                }
                else {
                    LiqGlyph        best_glyph= glyph;
                    LiqGlyphFlagSet best_flags= 0;
                    int             best_prior= 0;
                    for (; glyph; glyph= font->next_glyph(glyph)) {
                        LiqGlyphFlagSet flags= font->glyph_flags(glyph);
                        unless (flags & LIQ_GLYPH_ALT) { /* kerning alternatives are done later */
                            int prior= 0;

                            if (!(flags & LIQ_GLYPH_MONO) == !(font->mono))
                                prior+= 2;

                            if ((flags        & LIQ_GLYPH_JOIN_MASK) ==
                                (cur->joining & LIQ_JOIN_MASK_GLYPH))
                                prior+= 1;

                            if (prior > best_prior) {
                                best_prior= prior;
                                best_glyph= glyph;
                                best_flags= flags;
                            }
                        }
                    }
                    cur->code= best_glyph;
                }
            }
            else {
                /* character cannot be displayed, not even with a replacement
                 * char, so skip it. */
                cur->code= 0;
            }
        }

        cur->is_glyph= LIQ_TRUE;
        cur= nsm; /* skip over whole base + nsm* sequence */
    }

    /* character spacing and kerning: */
    unless (font->mono)
        liq_foreach_run (self, para, &liq_run_kerning);
}

static LiqX liq_compute_line_width (LiqRender *self, LiqLineInfo *line)
{
    LiqFont const *const font=  self->m_font;
    LiqCharInfo *const   first= line->first;
    LiqCharInfo *const   last=  line->last;
    LiqCharInfo *cur;
    LiqX width= 0;

    if (font->mono) {
        for (cur=first; cur < last; ++cur) {
            if (cur->code) {
                if (width == 0)
                    width += font->bitwidth;
                else
                    width += max_int(0, font->width + self->m_tracking);
            }
        }
    }
    else {
        for (cur=first; cur <= last; ++cur) {
            assert (cur->is_glyph);
            if (cur->code)
                width+= cur->kern + glyph_width (font, cur->code);
        }
    }

    return width;
}

static void liq_render_line (LiqRender *self, LiqLineInfo *line)
{
    LiqFont const *const   font=          self->m_font;
    void * const           bitmap=        self->m_bitmap;
    LiqBitmapClass * const bitmap_class= &self->m_bitmap_class;
    LiqCharInfo *const     first=         line->first;
    LiqCharInfo *const     last=          line->last;
    LiqColumnData          space_col=     0;
    LiqBool const          mono=          font->mono;
    LiqX const             xm=            font->bitwidth-1;
    LiqCharInfo *cur;
    LiqX x= line->pixel_start; /* current pixel pos */
    LiqX last_x= x;            /* beginning of the previous glyph */
    LiqColumnData any_pixel;

    for (cur=first; cur <= last; ++cur) {
        LiqGlyph const glyph= cur->code;
        if (glyph) {
            LiqX const x1= mono ? 0  : font->glyph_x1 (glyph);
            LiqX const x2= mono ? xm : font->glyph_x2 (glyph);
            LiqX xx;

            /* Draw a link from the current x position, maximally to the beginning of
             * the previous glyph. Stop when we hit a pixel. */
            if (space_col)
                for (xx= x; xx >= last_x; --xx)
                    if (bitmap_class->add (bitmap, xx, space_col) & space_col)
                        break;

            /* Update beginning of previous glyph */
            last_x= x;

            /* Insert next glyph */
            any_pixel= 0;
            if (cur->mirror) { /* mirror glyph */
                for (xx= x2; xx >= x1; --xx) {
                    LiqColumnData const col= font->glyph_data (glyph, xx);
                    bitmap_class->add (bitmap, x, col);
                    any_pixel|= col;
                    x++;
                }
            }
            else {
                for (xx= x1; xx <= x2; ++xx) {
                    LiqColumnData const col= font->glyph_data (glyph, xx);
                    bitmap_class->add (bitmap, x, col);
                    any_pixel|= col;
                    x++;
                }
            }

            /* Draw a link from previous x into current glyph if space_col
             * intersects with any pixel column */
            if (any_pixel & space_col)
                for (xx= last_x; xx < x; xx++)
                    if (bitmap_class->add (bitmap, xx, space_col) & space_col)
                        break;

            /* Kerning + new space_col */
            x= mono ? (last_x + font->width + self->m_tracking) : (x + cur->kern);

            space_col= font->glyph_space_col (glyph);
        }
    }
}

/* ********************************************************************** */
/* bitmap class for different types */

/* 8 bits */
static void liq_bitmap_set_8 (void *bitmap, LiqX x, LiqColumnData data)
{
    assert (x >= 0);
    ((LiqUInt8 *)bitmap)[x]= data;
}

static LiqColumnData liq_bitmap_add_8 (void *bitmap, LiqX x, LiqColumnData data)
{
    LiqColumnData old;
    assert (x >= 0);
    old= ((LiqUInt8 *)bitmap)[x];
    ((LiqUInt8 *)bitmap)[x]|= data;
    return old;
}

static LiqColumnData liq_bitmap_get_8 (void const *bitmap, LiqX x)
{
    assert (x >= 0);
    return ((LiqUInt8 const *)bitmap)[x];
}

static LiqBitmapClass const liq_bitmap_class_8= {
    &liq_bitmap_set_8,
    &liq_bitmap_add_8,
    &liq_bitmap_get_8,
};

/* 16 bits */
static void liq_bitmap_set_16 (void *bitmap, LiqX x, LiqColumnData data)
{
    assert (x >= 0);
    ((LiqUInt16 *)bitmap)[x]= data;
}

static LiqColumnData liq_bitmap_add_16 (void *bitmap, LiqX x, LiqColumnData data)
{
    LiqColumnData old;
    assert (x >= 0);
    old= ((LiqUInt16 *)bitmap)[x];
    ((LiqUInt16 *)bitmap)[x]|= data;
    return old;
}

static LiqColumnData liq_bitmap_get_16 (void const *bitmap, LiqX x)
{
    assert (x >= 0);
    return ((LiqUInt16 const *)bitmap)[x];
}

static LiqBitmapClass const liq_bitmap_class_16= {
    &liq_bitmap_set_16,
    &liq_bitmap_add_16,
    &liq_bitmap_get_16,
};

/* 32 bits */
static void liq_bitmap_set_32 (void *bitmap, LiqX x, LiqColumnData data)
{
    assert (x >= 0);
    ((LiqUInt32 *)bitmap)[x]= data;
}

static LiqColumnData liq_bitmap_add_32 (void *bitmap, LiqX x, LiqColumnData data)
{
    LiqColumnData old;
    assert (x >= 0);
    old= ((LiqUInt32 *)bitmap)[x];
    ((LiqUInt32 *)bitmap)[x]|= data;
    return old;
}

static LiqColumnData liq_bitmap_get_32 (void const *bitmap, LiqX x)
{
    assert (x >= 0);
    return ((LiqUInt32 const *)bitmap)[x];
}

static LiqBitmapClass const liq_bitmap_class_32= {
    &liq_bitmap_set_32,
    &liq_bitmap_add_32,
    &liq_bitmap_get_32,
};

/* ********************************************************************** */

void liq_render_clear_bitmap (LiqRender *self)
{
    TFREE(self->m_bitmap);
    self->m_width=  0;
}

static void liq_set_last_kern_0 (LiqCharInfo *first, LiqCharInfo *last)
{
    while (last >= first && !last->code) --last;
    if (last >= first) last->kern=0;
}

void liq_render_bitmap (LiqRender *self)
{
    int i;
    LiqX total_width;
    assert (self && self->m_info && "expected liq_render_set_text() first");
    assert (self && self->m_font && "expected liq_render_set_font() first");

    if (self->m_bitmap) /* already done? */
        return;

    /* in case no bidi() was used: */
    unless (self->m_bidi) {
        liq_split_lines (self);
        for (i=0; i < self->m_paragraph_cnt; i++) {
            liq_para_select_ligatures (self, self->m_paragraph + i);
            liq_para_select_glyphs    (self, self->m_paragraph + i);
        }
    }

    /* main rendering: */
    /* step 1: find line widths
     * One thing remains to be done: at the end of the (visual) line,
     * set the kerning back to 0.  We did not know before where this
     * would be (liq_run_kerning does not know which one is the
     * visually right-most char in a paragraph).
     */
    total_width= 0;
    for (i=0; i < self->m_line_cnt; i++) {
        liq_set_last_kern_0 (self->m_line[i].first, self->m_line[i].last);
        self->m_line[i].pixel_start= total_width;
        self->m_line[i].pixel_width= liq_compute_line_width (self, self->m_line + i);
        total_width+= self->m_line[i].pixel_width;
    }

    /* allocate */
    switch (self->m_font->height >> 3) {
    case 1:
        self->m_bitmap= (void*)TCALLOC(LiqUInt8, total_width);
        self->m_bitmap_class= liq_bitmap_class_8;
        break;

    case 2:
        self->m_bitmap= (void*)TCALLOC(LiqUInt16, total_width);
        self->m_bitmap_class= liq_bitmap_class_16;
        break;

    default:
        self->m_bitmap= (void*)TCALLOC(LiqUInt32, total_width);
        self->m_bitmap_class= liq_bitmap_class_32;
        break;
    }

    /* render */
    for (i=0; i < self->m_line_cnt; i++)
        liq_render_line (self, self->m_line + i);
}

